/*
 * COPYRIGHT:        See COPYRIGHT.TXT
 * PROJECT:          Ext2 File System Driver for WinNT/2K/XP
 * FILE:             except.c
 * PROGRAMMER:       Matt Wu <mattwu@163.com>
 * HOMEPAGE:         http://ext2.yeah.net
 * UPDATE HISTORY: 
 */

/* INCLUDES *****************************************************************/

#include "ntifs.h"
#include "ext2fs.h"

/* GLOBALS ***************************************************************/

extern PEXT2_GLOBAL Ext2Global;

/* DEFINITIONS *************************************************************/

#ifdef ALLOC_PRAGMA
#pragma alloc_text(PAGE, Ext2ExceptionFilter)
#pragma alloc_text(PAGE, Ext2ExceptionHandler)
#endif


NTSTATUS
Ext2ExceptionFilter (
             IN PEXT2_IRP_CONTEXT    IrpContext,
             IN PEXCEPTION_POINTERS ExceptionPointer)
{
    NTSTATUS Status, ExceptionCode;
    PEXCEPTION_RECORD ExceptRecord;

    ExceptRecord = ExceptionPointer->ExceptionRecord;

    ExceptionCode = ExceptRecord->ExceptionCode;

    Ext2BreakPoint();

    //
    // Check IrpContext is valid or not
    //

    if (IrpContext)
    {
        if ((IrpContext->Identifier.Type != EXT2ICX) ||
            (IrpContext->Identifier.Size != sizeof(EXT2_IRP_CONTEXT)))
        {
            IrpContext = NULL;
        }
    }
    else
    {
        if (FsRtlIsNtstatusExpected(ExceptionCode))
        {
            return EXCEPTION_EXECUTE_HANDLER;
        }
        else
        {
            Ext2BugCheck( EXT2_BUGCHK_EXCEPT, (ULONG_PTR)ExceptRecord,
                          (ULONG_PTR)ExceptionPointer->ContextRecord,
                          (ULONG_PTR)ExceptRecord->ExceptionAddress );
        }
    }

    //
    //  For the purposes of processing this exception, let's mark this
    //  request as being able to wait, and neither write through nor on
    //  removable media if we aren't posting it.
    //

    SetFlag(IrpContext->Flags, IRP_CONTEXT_FLAG_WAIT);

    if (FsRtlIsNtstatusExpected(ExceptionCode))
    {
        //
        // If the exception is expected execute our handler
        //

        Ext2Print((DBG_ERROR, "Ext2ExceptionFilter: Catching exception %xh\n",
            ExceptionCode));
        
        Status = EXCEPTION_EXECUTE_HANDLER;
        
        if (IrpContext)
        {
            IrpContext->ExceptionInProgress = TRUE;
            IrpContext->ExceptionCode = ExceptionCode;
        }
    }
    else
    {
        //
        // Continue search for an higher level exception handler
        //

        Ext2Print((DBG_ERROR, "Ext2ExceptionFilter: Passing on exception %#x\n",
            ExceptionCode));
        
        Status = EXCEPTION_CONTINUE_SEARCH;
        
        if (IrpContext)
        {
            Ext2FreeIrpContext(IrpContext);
        }
    }
    
    return Status;
}


NTSTATUS
Ext2ExceptionHandler (IN PEXT2_IRP_CONTEXT IrpContext)
{
    NTSTATUS Status;
    
    if (IrpContext)
    {
        if ( (IrpContext->Identifier.Type != EXT2ICX) || 
             (IrpContext->Identifier.Size != sizeof(EXT2_IRP_CONTEXT)))
        {
            Ext2BreakPoint();
            return STATUS_UNSUCCESSFUL;
        }

        Status = IrpContext->ExceptionCode;

        if (IrpContext->Irp)
        {
            //
            // Check if this error is a result of user actions
            //

            PIRP Irp = IrpContext->Irp;
        

            if (IoIsErrorUserInduced(Status))
            {
                //
                //  Now we will generate a pop-up to user
                //

                PDEVICE_OBJECT RealDevice;
                PVPB           Vpb = NULL;
                PETHREAD       Thread;

                if (IoGetCurrentIrpStackLocation(Irp)->FileObject != NULL)
                {
                    Vpb = IoGetCurrentIrpStackLocation(Irp)->FileObject->Vpb;
                }

                //
                // Get the initial thread
                //

                Thread = Irp->Tail.Overlay.Thread;
                RealDevice = IoGetDeviceToVerify( Thread );

                if ( RealDevice == NULL )
                {
                    //
                    // Get current thread
                    //

                    Thread = PsGetCurrentThread();
                    RealDevice = IoGetDeviceToVerify( Thread );

                    ASSERT( RealDevice != NULL );
                }

                if (RealDevice != NULL)
                {
                    //
                    //  Now we pop-up the error dialog ...
                    //

                    IoMarkIrpPending( Irp );
                    IoRaiseHardError( Irp, Vpb, RealDevice );

                    IoSetDeviceToVerify( Thread, NULL );

                    Status =  STATUS_PENDING;
                    goto errorout;
                }
            }

            IrpContext->Irp->IoStatus.Status = Status;
            
            Ext2CompleteRequest(IrpContext->Irp, FALSE, IO_NO_INCREMENT);
        }

errorout:
        
        Ext2FreeIrpContext(IrpContext);
    }
    else
    {
        Status = STATUS_INSUFFICIENT_RESOURCES;
    }
    
    return Status;
}

